#include <cpctelera.h>
#include "main.h"
#include "render.h"
#include "game.h"
#include "teclado.h"
#include "font_utils.h"
#include "objetos.h"
#include "gui.h"
#include "menu.h"
#include "memory.h"
#include "map_utils.h"
#include "tiles/tiles8x8.h"
#include "fonts/font4x5.h"
#include "sprites/bin/jugador_mago0.h"
#include "sprites/jugador_mago0.h"
#include "sprites/jugador_mago1.h"
// #include "sprites/hechizo.h"
#include "sprites/item_vida_rotate.h"
#include "sprites/item_hechizo_rotate.h"
#include "sprites/item_cambio_rotate.h"
#include "sprites/item_random_rotate.h"
#include "audio/m_ingame.h"
#include "audio/m_gameover.h"


// ==============================================================
// DEFINIR LA DESCRIPCIÓN DE LOS ITEMS RANDOM
// TAMBIÉN SE USAN PARA LA DESCRIPCIÓN DE LOS ITEMS NO RANDOM
// ==============================================================
const u8 txtItemRandom[13][15] = {
        "  +4 SPELLS   "
    ,   "   +1 LIFE    "
    ,   "CHARACTER SWAP"
    ,   "  FULL SPELLS " 
    ,   "   -1 LIFE    "
    ,   "   -1 SPELL   "
    ,   "PLAYER FREEZES"
    ,   "    SLOWER    "
    ,   "CRAZY PLAYER 1"
    ,   "CRAZY PLAYER 2"
    ,   "  FAST SPELLS "
    ,   "  +3 SPELLS   "
    ,   "PLAYER CHANGES"
};
 
// ==============================================================
// INICIALIZAR LOS VALORES DEL JUEGO
// ==============================================================
void init_game() {

    // Inicializar la fase
    initFase();

    // PINTAR EL MARCADOR
    pintarMarcadorSuperior();

    // REPRODUCIR LA MÚSICA
    //
    contador_musica = velocidad_musica = 5;
    cpct_akp_musicInit((u8*) m_ingame);
    reproducirMusica();

    // Activar el bit de partida en curso
    bitOptions = bitOptions | mask_playing_game;
    vaciaBufferTeclado();
}

// ==============================================================
// GESTIONAR LA CUENTA ATRÁS CUANDO ESTÁ ACTIVA
// SE LLAMARÁ A ESTA FUNCIÓN A TRAVÉS DE INTERRUPCIONES
// ==============================================================
void gameCountdown() {
    u8 x, color;
    if ((bitOptions & mask_enable_countdown) && (bitOptions & mask_playing_game)) {
        if (!countdownMin && !countdownSec) {
            backToMenu = SI;
            bitOptions = bitOptions - mask_playing_game;
        } else {
            countdownSec--;
            if (countdownSec > 60) {
                countdownSec = 59;
                countdownMin--;
                if (!countdownMin) {
                    printTxt(":", 39, Y_MARCADOR_TIME, NO, 3);
                }
            }
            if (!countdownMin) {
                color = 3;
            } else {
                color = 13;
            }
            x = 37;
            // y = Y_MARCADOR_TIME;
            printInt(countdownMin, x, Y_MARCADOR_TIME, NO, color);
            x += 4;
            if (countdownSec < 10) {
                printTxt("0", x, Y_MARCADOR_TIME, NO, color); 
                x += 2;
            }
            printInt(countdownSec, x, Y_MARCADOR_TIME, NO, color);
        }
    }
}

// ==============================================================
// Leer un objeto de la base de datos y asignarle sus propiedades
// Los objetos están definidos en el archivo ' objetos.h '
// Se leerán al inicio del nivel
//
//  - PARÁMETROS:
//      n:  El id del objeto a generar
// ==============================================================
void generarObjeto(u8 n) {

    // (TIPO < 2) GENERAL para los 2 jugadores
    if (objetos_ptr->tipo < 2) {
        jugador[n].x = jugador[n].x_anterior = jugador[n].x_anterior2 = jugador[n].x_origen = objetos_ptr->x;
        jugador[n].y = jugador[n].y_anterior = jugador[n].y_anterior2 = jugador[n].y_origen = objetos_ptr->y;
        jugador[n].movimientoH = jugador[n].movimientoV = jugador[n].estado = jugador[n].disparando = jugador[n].herido = jugador[n].respawn = 0;
        jugador[n].hechizos = HECHIZOS_INICIALES;
        jugador[n].vidas = VIDAS_INICIALES;
        jugador[n].mirando = jugador[n].direccion = mask_direccion_der; // De inicio miran los 2 hacia la derecha
        seleccionarSkin(n, n);

        hechizo[n].propiedades = 0;
    }

    // Forzamos el giro al jugador 1 para que mire a la izquierda de inicio
    jugador[1].direccion = mask_direccion_izq;

    // TIPO: (2) Objeto de tipo ITEM
    if (objetos_ptr->tipo == 2) {
        item[n-2].tipo = objetos_ptr->tipo;
        item[n-2].x = objetos_ptr->x;
        item[n-2].y = objetos_ptr->y;
        item[n-2].propiedades = 0; // Reseteamos sus propiedades
    }
}


// ==============================================================
// Sitúa el índide del objeto a leer justo en el objeto
// que corresponfa, teniendo en cuenta el 'mapa_actual'
// y leyéndolos de la base de datos de ' objetos.h '
// ==============================================================
void punteroObjetos() {
    u8 i;
    objetos_ptr = &objeto[0];
    i = 1;
    
    while (i < mapa_actual) {
        while (objetos_ptr->tipo != 255) { // El 9 es el valor clave que indica fin de los objetos de ese nivel
            objetos_ptr++;
        }
        objetos_ptr++; // Para saltar el valor clave de fin de objetos y dejar el puntero al inicio del siguiente objeto
        i++;
    }
}


// ==============================================================
// Leer los objetos de cada nivel y colocarlos
// Inicializar los valores por defecto antes de iniciar la fase_
// ==============================================================
void initFase() {
    u8 i;
    
    ///////////////////////////////
    /// Resetear los valores de los posibles items que hayan quedado de residuo de otras partidas
    ///
    for (i = 0; i < MAX_OBJETOS_ITEM; i++) {
        item[i].tipo = item[i].y = item[i].destruir = 0;
    }
    
    /////////////////////////////////////
    /// Seleccionar y pintar el mapa
    ///
    punteroMapa(NO);
    descomprimir_pintarMapa();
    
    //////////////////////////
    /// Colocar el puntero de objetos en la posición correspondiente al nivel actual
    ///
    punteroObjetos();
    
    //////////////////////////
    /// Leer los objetos del nivel y generarlos (hasta llegar al indicador de fin "255")
    ///
    i = 0;
    do {
        generarObjeto(i);
        objetos_ptr++;      // Avanzamos al siguiente objeto
        i++;                // Le asignamos el siguiente índice
    } while (objetos_ptr->tipo != 255); // El inidicador de fin de lista de objetos de ese nivel

    //////////////////////////
    /// Resetear los valores de las posibles hechizos que hayan quedado activos de residuo de otras partidas
    ///
    for (i = 0; i < MAX_OBJETOS_HECHIZO; i++) {
        hechizo[i].tipo = hechizo[i].destruir = hechizo[i].movimientoV = hechizo[i].movimientoH = hechizo[i].loops = 0; // De inicio están desactivados y parados
        hechizo[i].sprite_frame1 = (u8*) &sp_hechizo_1[0];
        hechizo[i].sprite_destruir = (u8*) &sp_hechizo_0[0];
    }

    //////////////////////////
    /// SWITCHES
    ///
    mapa_id_cambio = 0;
    reactivarSwitch();
    
    //////////////////////////
    /// ITEMS EN PANTALLA
    ///
    if (bitOptions & mask_item_en_pantalla) bitOptions = bitOptions - mask_item_en_pantalla;
    if (bitOptions & mask_player0_inverse) bitOptions = bitOptions - mask_player0_inverse;
    if (bitOptions & mask_player1_inverse) bitOptions = bitOptions - mask_player1_inverse;

    //////////////////////////
    /// REINICIAR VARIABLES
    ///
    anterior_id_item = 99; // Un número alto de un item que seguro que no exista
    countdownSec = ciclo = 0;
    countdownMin = 2;
    countdownInterrupt = 50;
}

// ==============================================================
//  El sonido que se reproducirá cuando un jugador se transforma
// ==============================================================
void sonidoTransforma() {
    playSFX(4,20);
}

// ==============================================================
//  El sonido que se reproducirá cuando se destruya un hechizo
// ==============================================================
void sonidoDestruyeHechizo() {
    playSFX(6,56);
}

// ==============================================================
// VERIFICAR SI UN JUGADOR COLISIONA CON ALGÚN ITEM
//
// - PARÁMETROS:
//      n: El id del item
//
// - DEVUELVE:
//      0:  Si no hay colisión
//      1:  Colisiona con el jugador 0 (0 +1)
//      2:  Colisiona con el jugador 1 (1 +1)
// ==============================================================
u8 hayColisionItem(u8 n) {
    u8 j;

    for (j = 0; j < NUM_JUGADORES; j++) {
        if ( jugador[j].x + SP_MAGO0_0_W > item[n].x && 
             jugador[j].x < item[n].x + SP_ITEM_VIDA_ROTATE_0_W &&
             jugador[j].y + SP_MAGO0_0_H > item[n].y && 
             jugador[j].y < item[n].y + SP_ITEM_VIDA_ROTATE_0_H ) {

                return j + 1; // Le sumamos 1 para que no sea 0. Después se lo restamos -\(._.)/-
        }

    }
    return 0;
}

// ==============================================================
// VERIFICAR SI UN JUGADOR COLISIONA CON ALGUN HECHIZO
//
// - PARÁMETROS:
//      n:  El id del hechizo a revisar (0 = del jugador 0 | 1 = del jugador 1)
//
// - DEVUELVE:
//      0:  Si no hay colisión
//      1:  Colisiona con el jugador 0 (0 +1)
//      2:  Colisiona con el jugador 1 (1 +1)
// ==============================================================
u8 hayColisionHechizo(u8 n) {
    u8 j;

    // Invertimos el ID del jugador ya que sólo lo comparamos con el jugador contrario
    j = n ^ 0b00000001;
    if  (   jugador[j].x + SP_MAGO0_0_W > hechizo[n].x && 
            jugador[j].x < hechizo[n].x + SP_HECHIZO_0_W &&
            jugador[j].y + SP_MAGO0_0_H > hechizo[n].y && 
            jugador[j].y < hechizo[n].y + SP_HECHIZO_0_H ) {
        
                return j + 1; // Le sumamos 1 para que no sea 0. Después se lo restamos -\(._.)/-
    }
    
    return 0;
}

// ==============================================================
// VERIFICAR SI LOS HECHIZOS 0 Y 1 COLISIONAN ENTRE SÍ
//
// - DEVUELVE:
//      0: No hay colisión entre hechizos
//      1: Hay colisión entre hechizos    
// ==============================================================
u8 hayColisionEntreHechizos() {
    if (    hechizo[0].y + SP_HECHIZO_0_H > hechizo[1].y && 
            hechizo[0].y < hechizo[1].y + SP_HECHIZO_0_H &&
            hechizo[0].x + SP_HECHIZO_0_W > hechizo[1].x && 
            hechizo[0].x < hechizo[1].x + SP_HECHIZO_0_W ) {
                return 1;
    }
    return 0;
}

// ==============================================================
// DECIDE QUÉ ACCIÓN SE REALIZA CUANDO SE RECOGE UN ITEM RANDOM (LOS DEL '?')
//
// - PARÁMETROS:
//      j: id del jugador que colisiona (0 o 1)
// ==============================================================
void accionesItemRandom(u8 j) {
    u8 i, p;
    i = cpct_rand() % 13;
    p = cpct_rand() % 2;

    // Cuando " i == 12 " se aplica la acción por defecto

    // ***** DEBUG *****
    // i = 2;
    // *****************

    switch( i ) {
        
    // +4 SPELLS
        case 0:     
            modificaMarcadorHechizos(j, 4);
            break;

    // +1_LIFE        
        case 1:     
            modificaMarcadorVida(j, 1);
            break;

    // CHARACTER SWAP        
        case 2:     
            if (jugador[0].skin == 0 && jugador[1].skin == 1) {
                // Si los 2 jugadores son el mago se fuerza a que sólo cambie 1 personaje con i = 12
                i = 12;
            } else {
                intercambiarPersonajes();
                sonidoTransforma();
            }
            break;

    // FULL SPELLS        
        case 3:     
            modificaMarcadorHechizos(j, 5);
            break;
        
    // -1 LIFE
        case 4:     
            modificaMarcadorVida(j, -1);
            break;
    
    // -1 SPELL
        case 5:     
            modificaMarcadorHechizos(j, -1);
            break;

    // SOMEONE FREEZE
        case 6:
            jugador[p].disparando = TIEMPO_DISPARANDO * 3;
            break;

    // FASTER        
        case 7:     
            jugador[j].velocidad++;
            break;
        
    // CRAZY PLAYER 1        
        case 8:     
            bitOptions = bitOptions | mask_player0_inverse;
            break;
        
    // CRAZY PLAYER 2        
        case 9:     
            bitOptions = bitOptions | mask_player1_inverse;
            break;
        
    // FAST SPELLS
        case 10:    
            // Si el disparo ya es el más rápido forzamos a otro efecto, cambio de personaje ( i = 12 )
            if (jugador[j].velocidadDisparo == 1) {
                i = 12;
            } else {
                --jugador[j].velocidadDisparo;
            }
            break;

    // +3 SPELL        
        case 11:    
            modificaMarcadorHechizos(j, 3);
            break;

    // SOMEONE_CHANGES        
        default:    
            i = 12;
    }

    // SOMEONE_CHANGES
    if (i == 12) {
        jugador[p].herido = TIEMPO_HERIDO * 2;
        sonidoTransforma();
    }

    printTxt(txtItemRandom[i], CENTRADO, 193, NO, 13);
}


// ==============================================================
// VERIFICAR LAS COLISIONES DE LOS JUGADORES CON LAS DEMÁS ENTIDADES
// Y REALIZAR LAS ACCIONES QUE CORRESPONDAN
// ==============================================================
void colisiones() {
    u8 i;
    u8 jugadorColisiona;

    //////////////////////////////////
    // COLISION CON ITEMS
    //////////////////////////////////
    //
    for (i = 0; i < MAX_OBJETOS_ITEM; i++) {
        if ((item[i].propiedades & mask_activo) && !(item[i].propiedades & mask_destruir)) {
            jugadorColisiona = hayColisionItem(i);
            
            if (jugadorColisiona) {
                jugadorColisiona--;

                playSFX(2, 55);

                item[i].propiedades = item[i].propiedades - mask_pintar;
                item[i].propiedades = item[i].propiedades | mask_destruir;
                item[i].destruir = 4;
                if (item[i].tipo == ITEM_HECHIZO) { 
                    modificaMarcadorHechizos(jugadorColisiona, 4);
                    printTxt(txtItemRandom[0], CENTRADO, 193, NO, 15);
                } else if (item[i].tipo == ITEM_VIDA) { 
                    modificaMarcadorVida(jugadorColisiona, 1);
                    printTxt(txtItemRandom[1], CENTRADO, 193, NO, 3);
                } else if (item[i].tipo == ITEM_CAMBIO) {
                    intercambiarPersonajes();
                    printTxt(txtItemRandom[2], CENTRADO, 193, NO, 9);
                } else  {
                    accionesItemRandom(jugadorColisiona);
                }
            }
        }
    }

    //////////////////////////////////
    // COLISION CON HECHIZOS
    //////////////////////////////////
    //
    for (i = 0; i < MAX_OBJETOS_HECHIZO; i++) {
        if (hechizo[i].propiedades & mask_hechizo_activo) {
            jugadorColisiona = hayColisionHechizo(i);
            if (jugadorColisiona && jugador[--jugadorColisiona].respawn == 0) {
                hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                hechizo[i].propiedades = hechizo[i].propiedades - mask_hechizo_activo;
                //playSFX(1, 35 + (jugadorColisiona * 5));
                modificaMarcadorVida(jugadorColisiona, -1);
                jugador[jugadorColisiona].herido = TIEMPO_HERIDO;
                sonidoTransforma();
                ++jugador[i].hits;
            }
        }
    }

    //if (hechizo[0].tipo && hechizo[1].tipo) { // TO TEST
        if ((hechizo[0].propiedades & mask_hechizo_activo) && (hechizo[1].propiedades & mask_hechizo_activo)) {
            if (hayColisionEntreHechizos()) {
                hechizo[0].destruir = hechizo[1].destruir = TIEMPO_DESTRUIR_HECHIZO;
                hechizo[0].propiedades = hechizo[1].propiedades = hechizo[1].propiedades - mask_hechizo_activo;
                sonidoDestruyeHechizo();
            }
        }
    //}

}

// ==============================================================
// ACCIONES QUE SE LANZARÁN CUANDO UN JUGADOR COLISIONE CON PICHOS
//
// - PARÁMETROS:
//      j: El id del jugador que colisiona con los pinchos (0 o 1)
// ==============================================================
void colisionPinchos(u8 j) {
    if (!jugador[j].respawn) {
        modificaMarcadorVida(j, -1);
        sonidoTransforma();
    }
    jugador[j].respawn = TIEMPO_RESPAWN;
}

// ==============================================================
// Mover las entidades por pantalla, 
//      verificar las colisiones con el escenario
// ==============================================================
void moverEntidades() {
    u8 i;
    i8 x;
    u8 colision, colisionTemp;

    // COLISIONES:
        // 0b00000001   -> INFERIOR
        // 0b00000010   -> IZQUIERDA
        // 0b00000100   -> DERECHA
        // 0b00001000   -> ESCALERA EN EL SUELO
        // 0b00010000   -> ESCALERA EN LA MITAD INFERIOR DEL JUGADOR


    // *****************************************
    // ENTIDADES JUGADOR
    // *****************************************
    //
    for (i = 0; i < NUM_JUGADORES; i++) {
        colision = colisionTemp = 0;

        // Los jugadores sólo se moverán si no están disparando ni han sido alcanzados por el hechizo enemiga
        if (!jugador[i].herido && !jugador[i].respawn) {

            ///////////////////////////////////////////////
            /// X - (SI) COINCIDE CON EL INICIO DE TILE
            ////////////////////////////////////////////////
            //
            // Sólo se necesita verificar las colisiones horizontales (izquierda y derecha)
            // cuando el sprite coincide con el inicio de un tile, ya que miden lo mismo de ancho
            //
            if ((jugador[i].x & 3) == 0) { // Lo mismo que " if (jugador[i].x % 4 == 0) "

                if ((jugador[i].y & 7) == 0) { // Lo mismo que " if (jugador[i].y % 8 == 0) "
                ///////////////////////////////////////////////
                /// X - (SI) COINCIDE CON EL INICIO DE TILE
                /// Y - (SI) COINCIDE CON EL INICIO DE TILE
                ////////////////////////////////////////////////
                //
                    // IZQUIERDA --> Colision con obstaculo
                    if (typeOfTile(tileValue(jugador[i].x - SP_MAGO0_0_W, jugador[i].y)) >= PLATAFORMA) {
                        colision = colision | mask_colision_izq;
                    } else if (typeOfTile(tileValue(jugador[i].x - SP_MAGO0_0_W, jugador[i].y + TL_TILESET_00_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_izq;
                    }

                    // DERECHA --> Colision con obstaculo
                    if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y)) >= PLATAFORMA) {
                        colision = colision | mask_colision_der;
                    } else if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + TL_TILESET_00_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_der;
                    }
                
                    // INFERIOR --> Colision con obstáculo o escalera
                    colisionTemp = typeOfTile(tileValue(jugador[i].x, jugador[i].y + SP_MAGO0_0_H));
                    if (colisionTemp == PINCHOS) { 
                        colisionPinchos(i);
                    } else if (colisionTemp >= PLATAFORMA) {
                        colision = colision | mask_colision_inferior;
                    } else if (colisionTemp == ESCALERAS) {
                        colision = colision | mask_escaleras_inferior;
                    }
                    
                    // MEDIA --> Colision con escalera
                    if (typeOfTile(tileValue(jugador[i].x, jugador[i].y + TL_TILESET_00_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    }
                    
                } else {
                ///////////////////////////////////////////////
                /// X - (SI) COINCIDE CON EL INICIO DE TILE
                /// Y - (NO) COINCIDE CON EL INICIO DE TILE
                ////////////////////////////////////////////////
                //

                    // IZQUIERDA --> Colision con obstaculo
                    if (typeOfTile(tileValue(jugador[i].x - TL_TILESET_00_W, jugador[i].y)) >= PLATAFORMA) {
                        colision = colision | mask_colision_izq;
                    } else if (typeOfTile(tileValue(jugador[i].x - TL_TILESET_00_W, jugador[i].y + TL_TILESET_00_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_izq;
                    } else if (typeOfTile(tileValue(jugador[i].x - TL_TILESET_00_W, jugador[i].y + SP_MAGO0_0_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_izq;
                    }

                    // DERECHA --> Colision con obstaculo
                    if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y)) >= PLATAFORMA) {
                        colision = colision | mask_colision_der;
                    } else if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + TL_TILESET_00_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_der;
                    } else if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + SP_MAGO0_0_H)) >= PLATAFORMA) {
                        colision = colision | mask_colision_der;
                    }

                    // MEDIA --> Colision con escaleras
                    if (typeOfTile(tileValue(jugador[i].x, jugador[i].y + SP_MAGO0_0_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    }

                }
            } else {

            ///////////////////////////////////////////////
            /// X - (NO) COINCIDE CON EL INICIO DE TILE
            ////////////////////////////////////////////////

                ///////////////////////////////////////////////
                /// X - (NO) COINCIDE CON EL INICIO DE TILE
                /// Y - (SI) COINCIDE CON EL INICIO DE TILE
                ////////////////////////////////////////////////
                //
                if ((jugador[i].y & 7) == 0) {
                    
                    // INFERIOR --> Colision con obstáculo o escalera (tile inferior izquierda)
                    colisionTemp = typeOfTile(tileValue(jugador[i].x, jugador[i].y + SP_MAGO0_0_H));
                    if (colisionTemp == PINCHOS) { 
                        colisionPinchos(i);
                    } else if (colisionTemp >= PLATAFORMA) {
                        colision = colision | mask_colision_inferior;
                    } else if (colisionTemp == ESCALERAS) {
                        colision = colision | mask_escaleras_inferior;
                    } 
                    
                    // INFERIOR --> Colision con obstáculo o escalera (tile inferior derecha)
                    colisionTemp = typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + SP_MAGO0_0_H));
                    if (colisionTemp == PINCHOS) { 
                        colisionPinchos(i);
                    } else 
                    if (colisionTemp >= PLATAFORMA) {
                        colision = colision | mask_colision_inferior;
                    } else if (colisionTemp == ESCALERAS) {
                        colision = colision | mask_escaleras_inferior;
                    }

                    // MEDIO --> Colision con escaleras enmedio
                    if (typeOfTile(tileValue(jugador[i].x, jugador[i].y + TL_TILESET_00_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    } else if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + TL_TILESET_00_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    }
                

                } else {
                ///////////////////////////////////////////////
                /// X - (NO) COINCIDE CON EL INICIO DE TILE
                /// Y - (NO) COINCIDE CON EL INICIO DE TILE
                ////////////////////////////////////////////////
                //
                    
                    // MEDIO --> Colision con escaleras enmedio
                    if (typeOfTile(tileValue(jugador[i].x, jugador[i].y + SP_MAGO0_0_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    } else if (typeOfTile(tileValue(jugador[i].x + TL_TILESET_00_W, jugador[i].y + SP_MAGO0_0_H)) == ESCALERAS) {
                        colision = colision | mask_escaleras_medio;
                    }
                }

            }


        ////////////////////////////////////////////
        /// FIJAR LAS COORDENADAS X e Y
        ////////////////////////////////////////////
        //
            // COLISIONES:
            // 0b00000001   -> INFERIOR
            // 0b00000010   -> IZQUIERDA
            // 0b00000100   -> DERECHA
            // 0b00001000   -> ESCALERA EN EL SUELO
            // 0b00010000   -> ESCALERA EN LA MITAD INFERIOR DEL JUGADOR

            // No hay colision inferior de obstáculo ni media de escalera ==> cae
            //
            if (!(colision & mask_colision_inferior) && !(colision & mask_escaleras_inferior) && !(colision & mask_escaleras_medio)) {
                jugador[i].y = jugador[i].y + VELOCIDAD_CAER; 
            } else {
            
                // Colision ESCALERAS sube y baja
                //
                if (jugador[i].movimientoV == ABAJO) {
                    if ((colision & mask_escaleras_medio) && !(colision & mask_colision_inferior) || (colision & mask_escaleras_inferior)) { 
                        // Según la velocidad del jugador comparando con el ciclo, el jugador se mueve o no
                        if (ciclo % jugador[i].velocidad == 0) {
                            jugador[i].y = jugador[i].y + VELOCIDAD_ESCALERAS;
                        }
                    }
                }
                if (jugador[i].movimientoV == ARRIBA) {
                    if (colision & mask_escaleras_medio) {
                        // Según la velocidad del jugador comparando con el ciclo, el jugador se mueve o no
                        if (ciclo % jugador[i].velocidad == 0) {
                            jugador[i].y = jugador[i].y - VELOCIDAD_ESCALERAS;
                        }
                    } 
                }
                
                /////////////////////////////////////////////
                // COLISION IZQUIERDA O DERECHA
                /////////////////////////////////////////////
                //
                
                x = jugador[i].x + jugador[i].movimientoH;
                
                if (!jugador[i].disparando ) {
                    if (jugador[i].movimientoH == IZQUIERDA) {
                        if (x < AREA_JUEGO_IZQ) {
                            jugador[i].x = 76;
                        } else 
                        if (!(colision & mask_colision_izq)) {
                            if (ciclo % jugador[i].velocidad == 0) {
                                jugador[i].x = x;
                            }
                        }
                    } else 
                    if (jugador[i].movimientoH == DERECHA) {
                        if ((x + SP_MAGO0_0_W) > AREA_JUEGO_DER) { 
                            // Como la x posicion X del jugador es sin signo, al intentar ser menor que cero
                            // la variable se desborda y vuelve a 255, por lo tanto será mayor que la zona
                            // de juego derecha.
                            // Así que lo devolvemos a = 0
                            //
                            jugador[i].x = 0;
                        } else 
                        if (!(colision & mask_colision_der) && ((jugador[i].x + SP_MAGO0_0_W) < AREA_JUEGO_DER)) {
                            if (ciclo % jugador[i].velocidad == 0) {
                                jugador[i].x = x;
                            }
                        }
                    }
                }
            }
        
        }
    }

    // *****************************************
    // ENTIDADES HECHIZO
    // *****************************************
    //
    for (i = 0; i < MAX_OBJETOS_HECHIZO; i ++) {
        if (hechizo[i].tipo != 0 && !hechizo[i].destruir) {
            
            // ********* MOVIMIENTO HORIZONTAL *********

            if (hechizo[i].propiedades & mask_hechizo_activo) {
                if (hechizo[i].movimientoH != PARADO) {
                
                    ///////////////////////////
                    // COLISION IZQUIERDA
                    ///////////////////////////
                    //
                    if (hechizo[i].movimientoH < 0) {

                        // BORDE DE LA PANTALLA DE JUEGO
                        if ((hechizo[i].x - VELOCIDAD_DISPARO_H) < AREA_JUEGO_IZQ) {
                            //if (bitOptions & mask_loop_spell) {
                                if (++hechizo[i].loops > MAX_LOOPS_HECHIZOS) {
                                    hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                    
                                } else {
                                    hechizo[i].x = AREA_JUEGO_DER - SP_HECHIZO_0_W;
                                }
                            //} else {
                            //    hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                            //}
                        } 


                            /////////////////////////////////////
                            // COLISION CON TILE OBSTACULO IZQUIERDA
                            //
                            // Solo cuando la x coincida con el inicio de un tile
                            if ((hechizo[i].x & 3) == 0) { // Lo mismo que " if (hechizo[i].x % 4 == 0) "
                                // 1 solo tile de alto
                                if (typeOfTile(tileValue(hechizo[i].x - SP_HECHIZO_0_W, hechizo[i].y)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                        
                                } else                        
                                // 2 tiles de alto
                                if ((hechizo[i].y & 7) > 1)  {
                                    if (typeOfTile(tileValue(hechizo[i].x - SP_HECHIZO_0_W, hechizo[i].y + SP_HECHIZO_0_H)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                        
                                    }
                                }
                            }
                        

                    } else {
                        
                        ///////////////////////////
                        // COLISION DERECHA
                        ///////////////////////////
                        //
                        if ((hechizo[i].x + SP_HECHIZO_0_W + VELOCIDAD_DISPARO_H) > AREA_JUEGO_DER) {
                            //if (bitOptions & mask_loop_spell) {
                                if (++hechizo[i].loops > MAX_LOOPS_HECHIZOS) {
                                    hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                    
                                } else {
                                    hechizo[i].x = 1;
                                }
                            //} else {
                                // hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                            //}
                        } else {

                            /////////////////////////////////////
                            // COLISION CON TILE OBSTACULO DERECHA
                            //
                            // Solo cuando la x coincida con el inicio de un tile
                            if ((hechizo[i].x & 3) == 0) {
                                // 1 solo tile de alto
                                if (typeOfTile(tileValue(hechizo[i].x + SP_HECHIZO_0_W, hechizo[i].y)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                        
                                } else 
                                // 2 tiles de alto (Sólo se verifica si en el IF anterio no se ha encontrado
                                //      un obstáculo y la posición del hechizo ocupa 2 tiles de alto)
                                if ((hechizo[i].y & 7) < 2)  {
                                    if (typeOfTile(tileValue(hechizo[i].x + SP_HECHIZO_0_W, hechizo[i].y + SP_HECHIZO_1_H)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                        
                                    }
                                }
                            }

                        }

                    }
                    // Movimiento del hechizo
                    if (!hechizo[i].destruir && !(ciclo % hechizo[i].velocidad)) {
                        hechizo[i].x = hechizo[i].x + hechizo[i].movimientoH;
                    }
                } else {

                    // ********* MOVIMIENTO VERTICAL *********
            
                    ///////////////////////////
                    // COLISION SUPERIOR
                    ///////////////////////////
                    //
                    if (hechizo[i].movimientoV < 0) {
                        
                        // BORDE DE LA PANTALLA DE JUEGO
                        if ((hechizo[i].y - VELOCIDAD_DISPARO_V ) < AREA_JUEGO_SUP) {
                            hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                            
                        } else {

                        /////////////////////////////////////
                        // COLISION CON TILE OBSTACULO SUPERIOR
                        /////////////////////////////////////

                            // X coincide con el inicio de un tile, revisamos un único tile
                            if (typeOfTile(tileValue(hechizo[i].x, hechizo[i].y - VELOCIDAD_DISPARO_V)) == OBSTACULO) {
                                    hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                    sonidoDestruyeHechizo();
                            } else if ((hechizo[i].x & 3) > 0) { // 2 tiles
                                if (typeOfTile(tileValue(hechizo[i].x + SP_HECHIZO_0_W, hechizo[i].y - VELOCIDAD_DISPARO_V)) == OBSTACULO) {
                                    hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                    
                                }
                            }
                        } 

                    } else {
                        ///////////////////////////
                        // COLISION INFERIOR
                        ///////////////////////////
                        //
                        // BORDE DE LA PANTALLA DE JUEGO
                        //if (hechizo[i].y % 2 == 0) {
                            if ((hechizo[i].y + SP_HECHIZO_0_H) > AREA_JUEGO_INF) {
                                hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                            } else {
    
                                // X coincide con el inicio de un tile, revisamos un único tile
                                if (typeOfTile(tileValue(hechizo[i].x, hechizo[i].y + SP_HECHIZO_0_H)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                } else if ((hechizo[i].x & 3) > 0) { // 2 tiles
                                    if (typeOfTile(tileValue(hechizo[i].x + SP_HECHIZO_0_W, hechizo[i].y + SP_HECHIZO_0_H)) == OBSTACULO) {
                                        hechizo[i].destruir = TIEMPO_DESTRUIR_HECHIZO;
                                    }
                                }
                            }
                        }
                    }
                    // Movimiento del hechizo
                    if (!hechizo[i].destruir && !(ciclo % hechizo[i].velocidad)) {
                        hechizo[i].y = hechizo[i].y + (hechizo[i].movimientoV * VELOCIDAD_DISPARO_V);
                    }
                }
            if (hechizo[i].destruir) {
                hechizo[i].propiedades = hechizo[i].propiedades - mask_hechizo_activo;
            }
        }
    }
}

// ==============================================================
// ACCIONES QUE SE REALIZAN CUANDO SE CREA UN ITEM NUEVO
// ==============================================================
void crearItem() {
    u8 k;
    u8 *sprite;

    // Elegimos un item al azar entre los del mapa.
    // Y tenemos en cuenta que no elija el mismo item 2 veces seguidas
    //
    if (!(bitOptions & mask_item_en_pantalla)) {

        bitOptions = bitOptions | mask_item_en_pantalla;

        do {
            k = cpct_rand() % MAX_OBJETOS_ITEM;
        } while (k == anterior_id_item); // while (item[k].y == 0 || k == anterior_id_item);

        // Guardamos el ID para hacer que el siguiente no sea el mismo
        anterior_id_item = k;

        // Si ninguno de los 2 jugadores tiene hechizos, forzamos que el item sea de hechizos
        if (jugador[0].hechizos == 0 || jugador[1].hechizos == 0) {
            item[k].tipo = ITEM_HECHIZO;
        } else {
        // Si los jugadores tienen hechizos, se selecciona un tipo de item al azar, pero descartando
        // el item de cambio de skin si los 2 jugadores tienen el skin por defecto (de mago)
            do {
                item[k].tipo = ((cpct_rand() & 0x03) + ITEM_SPAWNS_SPELL_PROBAB) + 1; // cpct_rand() % 4
            } while (item[k].tipo == ITEM_CAMBIO && (jugador[0].skin == SKIN_MAGO0 && jugador[1].skin == SKIN_MAGO1));
        }

        item[k].propiedades = mask_activo | mask_pintar;

        // ASIGNAR SPRITES AL ITEM
        if (item[k].tipo == ITEM_VIDA) {
            sprite = (u8*) &sp_item_vida_rotate_0[0];
        } else if (item[k].tipo == ITEM_CAMBIO) {
            sprite = (u8*) &sp_item_cambio_rotate_0[0];
        } else if (item[k].tipo == ITEM_RANDOM) {
            sprite = (u8*) &sp_item_random_rotate_0[0];
        } else {
            item[k].tipo = ITEM_HECHIZO;
            sprite = (u8*) &sp_item_hechizo_rotate_0[0];
        }
        
        //for (i = 0; i < 4; i++) {
            item[k].sprite_frame = sprite; // + (i * SP_ITEM_VIDA_ROTATE_0_W * SP_ITEM_VIDA_ROTATE_0_H);
        //}
   
        playSFX(3,70);
    }
}

// ==============================================================
// HACER QUE LOS JUGADORES SE INTERCAMBIEN EL PRESONAJE ENTRE SI
// SI UNO O LOS 2 JUGADORES ES EL MAGO NO LO INTERCAMBIA
// ==============================================================
void intercambiarPersonajes() {
    u8 n;
    
    girarSpritesSiMiraIzquierda();
    n = jugador[0].skin;
    jugador[0].skin = jugador[1].skin;
    jugador[1].skin = n;
    if (jugador[0].skin == SKIN_MAGO1) jugador[0].skin = SKIN_MAGO0;
    if (jugador[1].skin == SKIN_MAGO0) jugador[1].skin = SKIN_MAGO1;
    for (n = 0; n < NUM_JUGADORES; n++) {
        seleccionarSkin(n, jugador[n].skin);
        jugador[n].respawn = TIEMPO_RESPAWN_RECOLOCA;
        pintarCaraJugador(n);
    }
    sonidoTransforma();

    girarSpritesSiMiraIzquierda();
}

// ==============================================================
// ACCIONES A REALIZAR CUANDO UN JUGADOR DISPARE
// TIENE EN CUENTA QUE NO HAYA NINGÚN HECHIZO SUYO ACTIVO
//
// - PARÁMETROS:
//      j: El id del jugador que dispara (0 o 1)
// ==============================================================
void disparar(u8 j) {
    
    if (!(hechizo[j].propiedades & mask_hechizo_activo) && (!hechizo[j].destruir)) {
        if (jugador[j].hechizos > 0 && !jugador[j].disparando && !jugador[j].herido) {

            // Restar un hechizo
            jugador[j].hechizos--;
            
            // Sumar 1 al contador de disparos
            ++jugador[j].shots;
            // Activar el tiempo de pausa después del disparo
            jugador[j].disparando = TIEMPO_DISPARANDO;

            // Activar el hechizo
            hechizo[j].tipo = DISPARO_HECHIZO;
            hechizo[j].loops = 0;
            hechizo[j].propiedades = mask_hechizo_activo;
            
            /////////////////
            // VERTICAL
            /////////////////
            if (jugador[j].movimientoV != PARADO) {
                hechizo[j].velocidad = jugador[j].velocidadDisparo; // Multiplicar por movimientoV
                hechizo[j].x = jugador[j].x;
                hechizo[j].movimientoH = PARADO;
                hechizo[j].movimientoV = jugador[j].movimientoV;
                if (jugador[j].movimientoV == ARRIBA) {
                    hechizo[j].y = jugador[j].y;
                } else {
                    hechizo[j].y = jugador[j].y + SP_MAGO0_0_H;
                }

            } else {

            /////////////////
            // HORIZONTAL
            /////////////////
                hechizo[j].y = jugador[j].y + 8;    // Por defecto para los disparos horizontales
                hechizo[j].velocidad = jugador[j].velocidadDisparo;
                // HORIZONTAL
                if (jugador[j].mirando & mask_direccion_der) {
                    hechizo[j].x = jugador[j].x + 2;
                    hechizo[j].movimientoH = DERECHA;
                } else   {
                    hechizo[j].x = jugador[j].x - 2;
                    hechizo[j].movimientoH = IZQUIERDA;
                } 
            }
            
            hechizo[j].y_anterior = hechizo[j].y_anterior2 = hechizo[j].y;
            hechizo[j].x_anterior = hechizo[j].x_anterior2 = hechizo[j].x;

            playSFX(1, 49 + (j * 5));
            pintarIconosMarcador(j, ICONO_HECHIZO);
        }
    }
}

// ==============================================================
// CÁLCULOS EXTRA QUE SE REALIZARÁN EN CADA CICLO
// 
// - La cuenta atrás para destruir los hechizos
// - La cuenta atrás del jugador para el parón de después de disparar un hechizo
// - La cuenta atrás cuando el jugador reaparece después de transformarse
// - La cuenta atrás para destruir los items
// - La reactivación de los switches de los mapas que tengan
// ==============================================================
void calculos() {
    
    u8 i, a;
    
    ////////////////////
    // HECHIZOS
    ////////////////////
    //
    for (i = 0; i < MAX_OBJETOS_HECHIZO ; i++) {
        if (hechizo[i].destruir > 0) {
            if (hechizo[i].destruir == TIEMPO_DESTRUIR_HECHIZO) {
                sonidoDestruyeHechizo();
            }
            if (--hechizo[i].destruir == 0) {
                hechizo[i].tipo = hechizo[i].propiedades = 0;
                hechizo[i].movimientoH = hechizo[i].movimientoV = PARADO;
            }
        }
    }


    ////////////////////
    // JUGADORES
    ////////////////////
    //
    for (i = 0; i < NUM_JUGADORES; i++) {
        if (jugador[i].disparando) {
            jugador[i].disparando--;
        }
        if (jugador[i].herido) {
            if (jugador[i].herido == TIEMPO_HERIDO && jugador[i].vidas) {
                sonidoTransforma();
            }
            jugador[i].herido--;
            if (jugador[i].vidas) {
                if (jugador[i].herido == 1) {
                    // Si mira a la izquierda dejamos los sprites mirando a la derecha
                    if (jugador[i].mirando & mask_direccion_izq) {
                        girarSpritesJugador(i);
                    }
                    seleccionarSkin(i, SKIN_RANDOM);
                    pintarCaraJugador(i);
                    // Si mira a la izquierda giramos los sprites, que por defecto los habíamos dejado
                    //      siempre mirando a la derecha
                    if (jugador[i].mirando & mask_direccion_izq) {
                        girarSpritesJugador(i);
                    }
                }
            }
        }
        if (jugador[i].respawn) {
            jugador[i].respawn--;
            if (jugador[i].respawn == TIEMPO_RESPAWN_RECOLOCA) {
                jugador[i].x = jugador[i].x_origen;
                jugador[i].y = jugador[i].y_origen;
                sonidoTransforma();
            }
        }
    }




    ////////////////////
    // ITEMS
    ////////////////////
    //
    for (i = 0; i < MAX_OBJETOS_ITEM ; i++) {
        if (item[i].propiedades & mask_destruir) {
            item[i].destruir--;
            if (item[i].destruir == 0) {
                item[i].propiedades = 0;
                bitOptions = bitOptions - mask_item_en_pantalla;
            }
        }
    }

    // Hacer aparecer los items aleatoriamente
    if (ciclo == 1) {
        a = ITEM_SPAWN_BASE;
        if (!jugador[0].hechizos && !jugador[1].hechizos)
            a = ITEM_SPAWN_BASE_NOSPELLS;
        i = cpct_rand() % a;
        if (i < ITEM_SPAWNS_LOWER_THAN) crearItem();
    }

    ////////////////////
    // MAPA
    ////////////////////
    //
    if (mapa_halt_switch) {
        if (--mapa_halt_switch == 1) {
            // ****** DEBUG *****
                //printInt(mapa_halt_switch, 5, 192, NO, 0);
            // ****** DEBUG *****
            reactivarSwitch();
        }
    }

}


// ==============================================================
// BUCLE PRINCIPAL DEL JUEGO
// ==============================================================
void game() {

    // Inicializar el juego
    //
    while (opcionMenu == 2) {
        init_game();

        /////////////////////////
        // BUCLE DEL JUEGO
        /////////////////////////
        //
            while (!backToMenu && jugador[0].vidas && jugador[1].vidas) {

                bucle = 0;

                // Alternar entre los 2 bufers de video
                // Búfer 1: 0xC000
                // Búfer 2: 0x8000
                //
                    if (!vmem_swap) {
                        vmem_start = (u8*) VMEM_BUFFER2;
                        vmem_page = cpct_page80;
                    } else {
                        vmem_start = (u8*) VMEM_BUFFER1;
                        vmem_page = cpct_pageC0;
                    }
                //

                // cpct_setBorder(HW_RED);      // *** DEBUG
                colisiones();
                // cpct_setBorder(HW_BLUE);     // *** DEBUG
                teclado();
                // cpct_setBorder(HW_GREEN);    // *** DEBUG
                calculos();
                // cpct_setBorder(HW_YELLOW);   // *** DEBUG
                moverEntidades();
                // cpct_setBorder(HW_ORANGE);   // *** DEBUG
                borrarEntidades();
                // cpct_setBorder(HW_BLACK);    // *** DEBUG
                pintarEntidades();

                // Ralentizar todo para intentar que funcione a una velocidad constante
                while (bucle < 2)
                    cpct_waitVSYNC();

                cpct_waitVSYNC();
                cpct_setVideoMemoryPage(vmem_page); // Activar el búfer de video correspondiente
                vmem_swap = ~vmem_swap; // Cambiar el bit de alternar entre búfer de video con cada ciclo
                
                // A cada paso contamos un ciclo más, útil para las animaciones de sprites y para algunos cálculos
                ciclo++;
                if (ciclo == MAX_BUCLE_CICLOS) // No nos vayamos a pasar. Campo arbitrario.
                    ciclo = 0;
            }
        //////// FIN -- BUCLE DEL JUEGO   

        if (bitOptions & mask_playing_game) {
            bitOptions = bitOptions - mask_playing_game;
        }

        // Volvemos a dejar los sprites mirando a la derecha
        girarSpritesSiMiraIzquierda();

        // Mostrar la pantalla de game over
        gameOverScreen();
    }
} 