#include <cpctelera.h>
#include "render.h"
#include "gui.h"
#include "main.h"
#include "menu.h"
#include "teclado.h"
#include "font_utils.h"
#include "memory.h"
#include "sprites/bin/cara_mago0.h"
#include "sprites/cara_zombie.h"
#include "sprites/cara_mago0.h"
#include "sprites/cara_mago1.h"
#include "sprites/cara_rana.h"
#include "sprites/cara_larcenas.h"
#include "sprites/cara_bruja.h"
#include "sprites/muro_ladrillos.h"
#include "sprites/item_hechizo_rotate.h"
#include "sprites/item_vida_rotate.h"
#include "audio/m_gameover.h"


// ==============================================================
// SUMAR O RESTAR VIDA
//
// - PARÁMETROS:
//      j: El id del jugador a modificar
//      n: El número de vidas a sumar (si es positivo) o restar (si es negativo)
// ==============================================================
void modificaMarcadorVida(u8 j, i8 n) {
    jugador[j].vidas = ((jugador[j].vidas + n) > MAX_VIDAS ? MAX_VIDAS : jugador[j].vidas + n);
    pintarIconosMarcador(j, ICONO_VIDA);
}

// ==============================================================
// SUMAR O RESTAR HECHIZOS
//
// - PARÁMETROS:
//      j: El id del jugador a modificar
//      n: El número de hechizos a sumar (si es positivo) o restar (si es negativo)
// ==============================================================
void modificaMarcadorHechizos(u8 j, i8 n) {
    jugador[j].hechizos = ((jugador[j].hechizos + n) > MAX_HECHIZOS ? MAX_HECHIZOS : jugador[j].hechizos + n);
    if (jugador[j].hechizos < 0) jugador[j].hechizos = 0;
    pintarIconosMarcador(j, ICONO_HECHIZO);
}

// ==============================================================
// PINTA LA CARA DEL JUGADOR EN EL MARCADOR SUPERIOR
// Para ello primero la descomprime en la posición de memoria
// reservada ( cara_temp[] )
//
// - PARÁMETROS:
//      j: El id del jugador (0 es el izquierdo y 1 el derecho)
//      n: El id del personaje a dibujar, siendo:
//              0: Mago 0 (el izquierdo)
//              1: Mago 1 (el derecho)
//              2: Lárcenas
//              3: Rana
//              4: Zombie
//              5: Bruja
// ==============================================================
void pintarCaraJugador(u8 j) {
    
    u8 cara_temp[SP_CARA_MAGO0_SIZE];
    u8 *sprite;
    u8* vmem_actual; // Para el doble bufer
    u8* vmem_bufer = VMEM_BUFFER1;
    u8 i;
    
    sprite  = (u8*) cara_mago0_end;
    if (jugador[j].skin == SKIN_MAGO1) { sprite = (u8*) cara_mago1_end; } else
    if (jugador[j].skin == SKIN_LARCENAS) { sprite = (u8*) cara_larcenas_end; } else
    if (jugador[j].skin == SKIN_RANA) { sprite = (u8*) cara_rana_end; } else
    if (jugador[j].skin == SKIN_ZOMBIE) { sprite = (u8*) cara_zombie_end; } else 
    if (jugador[j].skin == SKIN_BRUJA) { sprite = (u8*) cara_bruja_end; }
    
    cpct_zx7b_decrunch_s((void *) (&cara_temp[0] + cara_mago0_size - 1), sprite);

    for (i = NUM_BUFFERS; i > 0; i--) {
        vmem_actual = cpct_getScreenPtr(vmem_bufer, X_CARA_PLAYER_1 + (j * (X_CARA_PLAYER_2 - X_CARA_PLAYER_1)), Y_CARA_PLAYERS);
        cpct_drawSprite(cara_temp, vmem_actual, SP_CARA_MAGO0_W, SP_CARA_MAGO0_H);
        vmem_bufer = VMEM_BUFFER2;
    }
    
    escribirNombreJugador(j);
}

// ==============================================================
// ESCRIBE EL NOMBRE DEL JUGADOR EN EL ESPACIO RESERVADO DE PANTALLA 
// EN EL MARCADOR SUPERIOR
// EL NOMBRE DEL PERSONAJE LO OBTIENE DEL ARRAY ' nomPersonajes[] '
//
// - PARÁMETROS:
//      j: El id del jugador (0 el izquierdo, 1 el derecho) 
//      n: El id del personaje a dibujar
// ==============================================================
void escribirNombreJugador(u8 j) {
    // Escribir su nombre en el marcador
    printTxt(nomPersonajes[jugador[j].skin], (j ? X_MARCADOR_PLAYER_2 : X_MARCADOR_PLAYER_1), Y_MARCADOR_SPELL + 1, NO, 0);
}

// ==============================================================
// PINTA LOS ICONOS DE HECHIZO O VIDA DEL MARCADOR SUPERIOR
//
// - PARÁMETROS:
//      j:      El id del jugador (0 el izquierdo, 1 el derecho) 
//      icono:  El icono a redibujadr (0 = Hechizos, 1 = Vidas)
// ==============================================================
void pintarIconosMarcador(u8 j, u8 icono) {
    u8 i, bufers, x, y;
    u8* vmem_actual;
    u8 *sprite;
    u8* vmem_bufer = (u8*) 0x8000;
    
    sprite = (u8*) &sp_corazon[0];
    y = Y_MARCADOR_LIFE;

    if (!icono) {
        sprite = (u8*) &sp_hechizo_0[0];
        y = Y_MARCADOR_SHOTS;
    }

    // PINTAR EL MARCADOR EN LOS 2 BUFERS DE VIDEO
    bufers = NUM_BUFFERS;
    do {
        // PINTAR TANTOS HECHIZOS COMO TENGA EL JUGADOR
        i = (icono ? jugador[j].vidas : jugador[j].hechizos);
        x = X_MARCADOR_PLAYER_1;
        vmem_actual = cpct_getScreenPtr(vmem_bufer, x + ((X_MARCADOR_PLAYER_2 - X_MARCADOR_PLAYER_1) * j), y);
        cpct_drawSolidBox (vmem_actual, 0, SP_HECHIZO_0_W * MAX_HECHIZOS, SP_HECHIZO_0_H);
        while (i > 0) {
            vmem_actual = cpct_getScreenPtr(vmem_bufer, x + (j == 0 ? ((MAX_HECHIZOS - i) * (SP_HECHIZO_0_W)) : (SP_HECHIZO_0_W * i)) + ((X_MARCADOR_PLAYER_2-X_MARCADOR_PLAYER_1) * j) - (SP_HECHIZO_0_W * j), y);
            cpct_drawSprite (sprite, vmem_actual, SP_HECHIZO_0_W, SP_HECHIZO_0_H);
            i--;
        }
        vmem_bufer = (u8*) 0xC000;
        bufers--;
    } while (bufers > 0);
}


// ==============================================================
// PINTAR UN MURO DE LADRILLOS (el sprite de ladrillos azules)
//
// - PARÁMETROS:
//      x:          Coordenada X superior izquierda
//      y:          Coordenada Y superior izquierda
//      filas:      Cantidad de filas de tiles
//      columnas:   Cantidad de columnas de tiles
// ==============================================================
void pintarMuro(u8 x, u8 y, u8 filas, u8 columnas) {
    u8 filas2, col2;
    u8 bufers = NUM_BUFFERS;
    u8* vmem_bufer;
    vmem_bufer = VMEM_BUFFER1;

    do {
        for (filas2 = 0; filas2 < filas; filas2++) {
            for (col2 = 0; col2 < columnas; col2++) {
                cpct_drawSprite (sp_muro_ladrillos, cpct_getScreenPtr(vmem_bufer, x + (col2 * SP_MURO_LADRILLOS_W), y + (filas2 * SP_MURO_LADRILLOS_H)), SP_MURO_LADRILLOS_W, SP_MURO_LADRILLOS_H);
            }
        }
        vmem_bufer = VMEM_BUFFER2;
    } while (bufers-- > 0);
}

// ==============================================================
// DIBUJA TODO EL MARCADOR SUPERIOR DEL JUEGO
// ==============================================================
void pintarMarcadorSuperior() {
    u8 i, j, x, k;
    u8* vmem_bufer;
    u8* position;

    pintarMuro(0, 0, 4, 20);
    vmem_bufer = VMEM_BUFFER1;

    for (i = NUM_BUFFERS; i > 0; i--) {
        
        // Fondo azul claro
        position = cpct_getScreenPtr(vmem_bufer, X_MARCADOR_PLAYER_1-1, Y_MARCADOR_SHOTS - 1);
        cpct_drawSolidBox (position, cpct_px2byteM0(6,6), 54, 27);
        
        // Recuadro negro de SHOTS - LIFE - PLAYER
        x = X_MARCADOR_PLAYER_1;
        for (j = 0; j < 2; j++) {
            position  = cpct_getScreenPtr(vmem_bufer, x, Y_MARCADOR_SPELL);
            cpct_drawSolidBox (position, 0, SP_HECHIZO_0_W * MAX_HECHIZOS, SP_HECHIZO_0_H);
            x = X_MARCADOR_PLAYER_2;
        }
        
        // Fondo Textos centrales
        for (k = Y_MARCADOR_SHOTS; k <= Y_MARCADOR_SPELL; k = k + 9) {
            position = cpct_getScreenPtr(vmem_bufer, 35, k);
            cpct_drawSolidBox (position, cpct_px2byteM0(1,1), 10, 7);
        }

        // Recuadro TIME
        if (bitOptions & mask_enable_countdown) {
            position = cpct_getScreenPtr(vmem_bufer, 34, Y_MARCADOR_TIME - 1);
            cpct_drawSolidBox (position, cpct_px2byteM0(6,6), 12, 8);
            // Recuadro Negro
            position  = cpct_getScreenPtr(vmem_bufer, 35, Y_MARCADOR_TIME - 1);
            cpct_drawSolidBox (position, 0, 10, 7);
            // Los 2 puntos
            printTxt(":", 39, Y_MARCADOR_TIME, NO, 13);
        }

        vmem_bufer = VMEM_BUFFER2;
    }

    for (i = 0; i < NUM_JUGADORES; i++) {
        pintarCaraJugador(i);
        pintarIconosMarcador(i, ICONO_VIDA);
        pintarIconosMarcador(i, ICONO_HECHIZO);
    }
    
    printTxt("SHOTS", CENTRADO, Y_MARCADOR_SHOTS + 1, SI, 0);
    printTxt("LIVES", CENTRADO, Y_MARCADOR_LIFE + 1, SI, 0);
    printTxt("NAME", CENTRADO, Y_MARCADOR_SPELL + 1, SI, 0);

}

void drawStats(u16 num, u8 stat, u8 player) {
    u8 x = 6 + (player * 61);
    u8 y = 79;
    u8 x2 = 0;
    u8 txt[6];
    
    switch (stat) {
        case 0:
            cpct_memcpy (txt, "WINS\0", 5);
            break;
        case 1:
            cpct_memcpy (txt, "SHOTS\0", 6);
            x2 = 1;
            break;
        case 2:
            cpct_memcpy (txt, "HITS\0", 5);
            break;
        /*
        case 3:
            cpct_memcpy (txt, "RATE\0", 5);
            break;
        */
    }
    
    y += stat * 23;
    printTxt(txt,  x - x2,  y, NO, 10);
    printTxt("00", x + 1,   y + 9, NO, 13);
    if (num > 99)      x += 1;
    else if (num > 9)  x += 3;
    else               x += 5;
    printInt(num,  x, y + 9, NO, 13);
    
}
    

// ==============================================================
// DIBUJA LA PANTALLA DE GAME OVER
// ==============================================================
void gameOverScreen() {
    u8 playerWin, x, y, i;
    u8 color, byteColor;
    u16 timeWait;
    
    playerWin = 0; // opcionMenu = 0;
    y = 72;

    cpct_akp_musicInit((u8*) m_gameover);
    reproducirMusica();

    // REVISAR EL JUGADOR GANADOR
        if (bitOptions & mask_enable_countdown) {
            while (jugador[0].vidas > 0 && jugador[1].vidas > 0) {
                --jugador[0].vidas;
                --jugador[1].vidas;
            }
        }
        if (!jugador[0].vidas || !jugador[1].vidas) {
            pintarRecuadroDoble (19, y, 42, 34);
            if (jugador[0].vidas != jugador[1].vidas) {
                printTxt("WINNING PLAYER", 24, y + 23, NO, 13);
                if (jugador[0].vidas == 0) {
                    playerWin = 2;
                } else {
                    playerWin = 1;
                }
                //playerWin = (jugador[0].vidas ? 1 : 2); // playerWin = 1;
                printInt(playerWin, 54, y + 23, NO, 13);
                jugador[playerWin - 1].wins++;
            } else {
                printTxt("TIE", CENTRADO, y + 23, NO, 13);
            }
        } else {
            pintarRecuadroDoble (19, y, 42, 19);
        }

        printTxt("GAME OVER", CENTRADO, y + 7, NO, 10);

        // VENTANA DE ESTADÍSTICAS DE JUGADORES
        x = 3;
        for (i = 0; i < NUM_JUGADORES; i++) {

            if (jugador[i].wins > 253)  jugador[i].wins  = 1; /// OMG!! QUIÉN LLEGARÁ A JUGAR 254 PARTIDAS??
            if (jugador[i].shots > 999) jugador[i].shots = 1; /// Y A 1000 SHOTS ??
            if (jugador[i].hits > 999)  jugador[i].hits  = 1; /// Y YA 1000 HITS ....
            
            pintarRecuadroDoble(x, y, 13, 74);

            drawStats(jugador[i].wins, 0, i);
            drawStats(jugador[i].shots, 1, i);
            drawStats(jugador[i].hits, 2, i);

            x = 64;
        }

        // VENTANA DE TECLAS
        pintarRecuadroDoble (19, y+50, 42, 44);
        printTxt("PRESS", CENTRADO, y+57, NO, 8);
        printTxt("/R/ TO RESTART", CENTRADO, y+73, NO, 0);
        printTxt("/M/ TO MENU", CENTRADO, y+81, NO, 0);
        
        opcionMenu = 0;
        color = 9;
        timeWait = 0;
        vaciaBufferTeclado();
        y = 72;
        do {
            cpct_scanKeyboard();

            if (cpct_isKeyPressed(Key_M)) {
                opcionMenu = VOLVER_AL_MENU;
            }
            if (cpct_isKeyPressed(Key_R)) {
                opcionMenu = SEGUIR_JUGANDO;
                backToMenu = NO;
                vaciaBufferTeclado();
            }

            /// PARPADEAR EL MARCADOR DEL GANADOR
            if (playerWin != 0) {
                if (++timeWait > 1200) {
                    u8* vmem_pos;
                    color = (color == 9 ? 14 : 9);
                    byteColor = cpct_px2byteM0(color, color);
                    x = 3;
                    if (playerWin == 2) x = 64;
                    vmem_pos = cpct_getScreenPtr (vmem_start, x, 72);
                    cpct_drawSolidBox (vmem_pos, byteColor, 13, 2);
                    cpct_drawSolidBox (vmem_pos, byteColor, 1, y + 2);
                    cpct_drawSolidBox (cpct_getScreenPtr (vmem_start, x, 72 + y), byteColor, 13, 2);
                    cpct_drawSolidBox (cpct_getScreenPtr (vmem_start, x + 12, 72), byteColor, 1, y + 2);
                    timeWait = 0;
                }
            }
            ///
       } while(!opcionMenu);
    sonidoTecla();       
}