////////////////////////////////////
// BE WITCHED OF BE DEAD
//
// by HINAMAN 2022
//
// Released under GNU GENERAL PUBLIC LICENSE v3
//
#include <cpctelera.h>
#include "main.h"
#include "sprites/paleta.h"
#include "tiles/tiles8x8.h"
#include "font_utils.h"
#include "menu.h"
#include "render.h"
#include "game.h"
#include "teclado.h"
#include "memory.h"
#include "audio/m_sfx.h"
#include "audio/m_bewitch.h"

u8 ciclo;
u8 mapa_actual;
u8 selectedStage;
u16 mapa_halt_switch;
u8 mapa_id_cambio;
u8 mapa[TILES_TOTALES];
u8* vmem_start;
u8 vmem_page;
u8 vmem_swap;
u8* mapa_ptr;
u8 backToMenu;
u8 bucle;
u8 contador_musica;
u8 velocidad_musica;
u8 opcionMenu;
u8 bitOptions;
u8 anterior_id_item;
u8 countdownSec;
u8 countdownMin;
u8 countdownInterrupt = 1;

TKey key[NUM_JUGADORES];
TJugador jugador[NUM_JUGADORES];
THechizo hechizo[MAX_OBJETOS_HECHIZO];
TItem item[MAX_OBJETOS_ITEM];
TObjeto* objetos_ptr;

const u8 sprites_player0[384];
const u8 sprites_player1[384];

// ==============================================================
// DEFINIR LOS NOMBRES DE LOS PERSONAJES
// ==============================================================
const u8 nomPersonajes[6][11] = {
      "ED SPELLS "
   ,  " MAGIK MAX"
   ,  " LARCENAS "
   ,  "  FROGGY  "
   ,  "  ZOMBIE  "
   ,  " WITCHER  "
};

// ==============================================================
// DEFINIR LAS TECLAS POR DEFECTO
// ==============================================================
void setDefaultKeys() {
   key[0].up    = Key_1;
   key[0].down  = Key_Q;
   key[0].left  = Key_D;
   key[0].right = Key_F;
   key[0].fire  = Key_Z;

   key[1].up    = Key_F7;
   key[1].down  = Key_F4;
   key[1].left  = Key_F2;
   key[1].right = Key_F3;
   key[1].fire  = Key_F0;
}

// ==============================================================
// CREAR TABLA DE TRANSPARENCIAS
// ==============================================================
cpctm_createTransparentMaskTable(sp_tablaTrans, 0x200, M0, 0);

// ==============================================================
// INTERRUPCIONES
// ==============================================================
void setInterruptAtVSYNCStart(void *intHdl) {
    cpct_waitVSYNC();
    __asm
        halt
        halt
    __endasm;
    cpct_waitVSYNC();
    cpct_setInterruptHandler(intHdl);
}

// ==============================================================
// MANEJADOR DE INTERRUPCIONES
// ==============================================================
void intHandler() {
   if(--contador_musica == 0){
      contador_musica = velocidad_musica + 1;
      playMusic();
   }
   if (--countdownInterrupt == 0) {
      countdownInterrupt = 254; // 5 interrupciones por segundo aprox
      gameCountdown();
   }

   bucle++;
}

void initComputer() {
   cpct_disableFirmware();
   setInterruptAtVSYNCStart(intHandler);
   cpct_setBorder(HW_BLACK);
   cpct_setPalette(paleta, 16);
   cpct_setVideoMode(0);
}


// ==============================================================
// FUNCIÓN PRINCIPAL MAIN
// ==============================================================
void main(void) {
   cpct_setStackLocation ((u8*) 0x7FFF);
   initComputer();

   bitOptions = mask_sfx_on | mask_music_on;
   selectedStage = 1;

   setDefaultKeys();
   initSprites();

   while(1) {
      menu();
      backToMenu = 0;
      while (opcionMenu == SEGUIR_JUGANDO) {
         game();
      }
   }
}

// ==============================================================
// PONER LAS 2 PANTALLAS DEL BUFFER EN NEGRO
// ==============================================================
void borraPantallas() {
    cpct_memset(VMEM_BUFFER1, 0, 0x4000);
    cpct_memset(VMEM_BUFFER2, 0, 0x4000);
}

// ==============================================================
// ESTABLECER EL BUFFER 1 COMO BUFER DE VIDEO
// ==============================================================
void seleccionarMonoBufer() {
   vmem_swap   = 0;
   vmem_start  = VMEM_BUFFER1;
   vmem_page   = cpct_pageC0;
   cpct_setVideoMemoryPage(vmem_page);
}

// ==============================================================
// LLAMAR A LA REPRODUCCIÓN DE MÚSICA EN LAS INTERRUPCIONES
// ==============================================================
void playMusic() {
   __asm
      .db 0x08
      EXX
      PUSH AF
      PUSH BC
      PUSH DE
      PUSH HL
   __endasm;
   cpct_akp_musicPlay();
   __asm
      POP HL
      POP DE
      POP BC
      POP AF
      EXX
      .db 0x08
   __endasm;
}

// ==============================================================
// REPRODUCIR UN EFECTO DE SONIDO
// SI LOS SFX ESTÁN DESACTIVADOS NO REPRODUCE NADA
//
// - PARÁMETROS:
//    insturmento:   El id del instrumento, dentro de la canción de Arkos Tracker (el primero es el 1)
//    nota:          Nota a reproducir del instrumento anterior
// ==============================================================
void playSFX(u8 instrumento, u8 nota) {
   if (bitOptions & mask_sfx_on) {
      cpct_akp_SFXInit (m_sfx);
      cpct_akp_SFXPlay (instrumento, 14, nota, 0, 0, AY_CHANNEL_B);
   }
}

// ==============================================================
// REPRODUCIR UNA CANCIÓN DE ARKOS TRACKER
// SI LA MÚSICA ESTÁ DESACTIVADA REPRODUCE LA PISTA DE SFX (EQUIVALE A SILENCIO)
//
// - REQUISITOS:
//    La canción a reproducir se debe haber definido previamente con ' cpct_akp_musicInit( ) '
// ==============================================================
void reproducirMusica() {
    if (!(bitOptions & mask_music_on)) {
        cpct_akp_musicInit((u8*) m_sfx);
    }
    cpct_akp_musicPlay();
}