#include <cpctelera.h>
#include "render.h"
#include "main.h"
#include "gui.h"
#include "tiles/tiles8x8.h"
#include "fonts/font4x5.h"
#include "font_utils.h"
#include "memory.h"
#include "sprites/item_vida_rotate.h"
#include "sprites/efecto_transforma.h"
//#include "sprites/hechizo.h"
#include "sprites/jugador_mago0.h"
#include "sprites/jugador_mago1.h"
#include "sprites/jugador_larcenas.h"
#include "sprites/jugador_rana.h"
#include "sprites/jugador_zombie.h"
#include "sprites/jugador_bruja.h"
#include "sprites/bin/jugador_mago0.h"
#include "sprites/bin/jugador_mago1.h"
#include "sprites/bin/jugador_larcenas.h"
#include "sprites/bin/jugador_rana.h"
#include "sprites/bin/jugador_zombie.h"
#include "sprites/bin/jugador_bruja.h"
#include "sprites/cara_mago0.h"
#include "sprites/cara_mago1.h"
#include "sprites/cara_larcenas.h"
#include "sprites/cara_rana.h"
#include "sprites/cara_zombie.h"
#include "map_utils.h"
#include "maps/mapa1.h"

// ==============================================================
// INICIZALIZAR LOS SPRITES DE LOS 2 JUGADORES CON LOS PUNTEROS
// A LAS POSICIONES DE LOS SPRITES
// Y INICIALIZAR LOS VALORES DE INICIO
// ==============================================================
void initSprites() {
    u8 *sprite;
    u8 i;

    sprite = (u8*) &sprites_player0;
    for (i = 0; i < NUM_JUGADORES; i++) {
        // Fijar los punteros de los sprites de jugadores
        jugador[i].skin                 = i;
        jugador[i].sprite_frame[0]      = sprite;
        jugador[i].sprite_frame[1]      = sprite +  SP_MAGO0_0_SIZE;
        jugador[i].sprite_escala_frame1 = sprite + (SP_MAGO0_0_SIZE *2);
        jugador[i].sprite_escala_frame2 = sprite + (SP_MAGO0_0_SIZE *3);
        jugador[i].sprite_disparando    = sprite + (SP_MAGO0_0_SIZE *4);
        jugador[i].sprite_herido        = sprite + (SP_MAGO0_0_SIZE *5);
        jugador[i].wins = jugador[i].hits = jugador[i].shots = 0;
        
        sprite = (u8*) &sprites_player1;
    }
}

// ==============================================================
// ASIGNAR UN SKIN DE PERSONAJE A UN JUGADOR
//
// - PARÁMETROS:
//      j: id del jugador al que asignamos el skin
//      n: id del skin (personaje) a asignar, siendo:
//              (0): Mago 0 (Izquierda)
//              (1): Mago 1 (Derecha)
//              (2): Lárcenas
//              (3): Rana
//              (4): Zombie
//              (5): Bruja
// ==============================================================
void seleccionarSkin(u8 j, u8 n) {
    u8 *sprite;
    u8 *sprite_descomprimido;

    // Seleccionar personaje al azar (teniendo en cuenta que el jugador 0 no puede ser el mago 1 
    //    y viceversa) y que tampoco seleccionaremos 2 jugadores iguales (por el problema de girar los sprites)
    //    ni el mismo personaje que ya es actualmente
    if (n == SKIN_RANDOM) {
        do {
            n = cpct_rand() % SKINS_DIFERENTES;
        } while ((n == 0 && j == 1) || (n == 1 && j == 0) || n == jugador[0].skin || n == jugador[1].skin);
    } 
    jugador[j].skin = n;

    // DEFINIR SPRITES Y CARACTERISTICAS DE CADA PERSONAJE
        // Por defecto
        sprite = (u8*) jugador_mago0_end;
        jugador[j].velocidad = VEL_RETARDO_MAGO;
        jugador[j].velocidadDisparo = 2;

    if (n == 1) { 
        sprite = (u8*) jugador_mago1_end; 
    } else
    if (n == 2) { 
        sprite = (u8*) jugador_larcenas_end; 
        jugador[j].velocidad = VEL_RETARDO_LARCENAS;
    } else
    if (n == 3) { 
        sprite = (u8*) jugador_rana_end; 
        jugador[j].velocidad = VEL_RETARDO_RANA;
    } else
    if (n == 4) { 
        sprite = (u8*) jugador_zombie_end; 
        jugador[j].velocidad = VEL_RETARDO_ZOMBIE;
        jugador[j].velocidadDisparo = 1;
    } else 
    if (n == 5) { 
        sprite = (u8*) jugador_bruja_end; 
        jugador[j].velocidad = VEL_RETARDO_BRUJA;        
    }
    
    // Descomprimir los sprites
    //
    if (j == 0) {
        sprite_descomprimido = (u8*) sprites_player0;
        if (bitOptions & mask_player0_inverse) bitOptions = bitOptions - mask_player0_inverse;
    } else {
        sprite_descomprimido = (u8*) sprites_player1;
        if (bitOptions & mask_player1_inverse) bitOptions = bitOptions - mask_player1_inverse;
    }
    
    cpct_zx7b_decrunch_s((void *) (sprite_descomprimido + jugador_mago0_size - 1), (u8*) sprite);

}

// ==============================================================
// VOLTEAR HORIZONTALMENTE EL SPRITE DE UN PERSONAJE
// PARA CUANDO CAMBIA DE DIRECCIÓN ENTRE IZQUIERDA O DERECHA
//
// - PARÁMETROS:
//      j: El id del jugador que queremos voltear
// ==============================================================
void girarSpritesJugador(u8 j) {
    cpct_hflipSpriteM0(SP_MAGO0_0_W, SP_MAGO0_0_H, jugador[j].sprite_frame[0]); 
    cpct_hflipSpriteM0(SP_MAGO0_0_W, SP_MAGO0_0_H, jugador[j].sprite_frame[1]);
    cpct_hflipSpriteM0(SP_MAGO0_0_W, SP_MAGO0_0_H, jugador[j].sprite_disparando);
}

// ==============================================================
// VERIFICAR SI ALGUNO DE LOS 2 JUGADORES ESTÁ MIRANDO A LA IZQUIERDA
// PARA VOLTEAR SU SPRITE Y DEJARLOS POR DEFECTO MIRANDO A LA DERECHA
// ==============================================================
void girarSpritesSiMiraIzquierda() {
    u8 n;
    for (n = 0; n < NUM_JUGADORES; n ++) {
        if (jugador[n].mirando & mask_direccion_izq) {
            girarSpritesJugador(n);
        }
    }
}

// ==============================================================
// BORRAR LAS ENTIDAD EN PANTALLA REDIBUJANDO ENCIMA SUYO LOS TILES QUE OCUPA
// ==============================================================
void borrarEntidades() {
    u8 i, w, h;

    // ENTIDADES JUGADOR
    //
    for (i = 0; i < NUM_JUGADORES; i++) {
        w = 1 + (jugador[i].x_anterior2 % TL_TILESET_00_W ? 1 : 0); // Dibujar 1 tile y en caso de estar en posición impar: dibujar una mas
        h = 2 + (jugador[i].y_anterior2 % TL_TILESET_00_H ? 1 : 0); // Dibujar 2 tiles y en caso de estar en la posición 3, dibujar una mas, en caso contrario: 0
        redibujarTile(jugador[i].x_anterior2, jugador[i].y_anterior2, w, h, SI);
    }

    // ENTIDADES HECHIZO
    //
    for (i = 0; i < MAX_OBJETOS_HECHIZO; i++) {
        if (hechizo[i].tipo != 0) {
            w = 1 + (hechizo[i].x_anterior2 % TL_TILESET_00_W ? 1 : 0);
            h = 1 + ((hechizo[i].y_anterior2 % TL_TILESET_00_H > 1) ? 1 : 0);
            redibujarTile(hechizo[i].x_anterior2, hechizo[i].y_anterior2, w, h, SI);
        }
    }
    
    // ENTIDADES ITEM
    //
    // No necesitamos borrarlas, los dibujamos sin transparencia y ya se borran solos
    /*
    for (i = 0; i < MAX_OBJETOS_ITEM; i++) {
        if (item[i].propiedades & mask_activo) {
            redibujarTile(item[i].x, item[i].y, 1, 2, SI);
        }
    }
    */
}

// ==============================================================
// PINTAR LAS ENTIDADES EN PANTALLA (EN EL BUFER ACTIVO)
// ==============================================================
void pintarEntidades() {
    u8 i, frame;
    u8 *sprite;

    /////////////////////////////////////////////////////
    // ENTIDADES JUGADOR
    /////////////////////////////////////////////////////
    // 
    for (i = 0; i < NUM_JUGADORES; i++) {

        // GIRAR EL SPRITE SI LA DIRECCIÓN ES DIFERENTE HACIA DONDE MIRA
        //
        if (jugador[i].direccion != jugador[i].mirando) {
            girarSpritesJugador(i);
            jugador[i].mirando = jugador[i].direccion;
        }

        // CAMBIAR DE FRAME DE LA ANIMACIÓN SEGÚN EL CICLO
        //

        if (jugador[i].respawn) {
            sprite = jugador[i].sprite_herido;
            
            if (jugador[i].respawn < TIEMPO_RESPAWN_GIRA4) {
                    sprite = (u8*) sp_efecto_transforma_1;
            } else if (jugador[i].respawn < TIEMPO_RESPAWN_GIRA3) {
                    sprite = (u8*) sp_efecto_transforma_0;
            } else if (jugador[i].respawn < TIEMPO_RESPAWN_RECOLOCA) {
                    sprite = (u8*) sp_efecto_transforma;
            } else if (jugador[i].respawn < TIEMPO_RESPAWN_GIRA2) {
                    sprite = (u8*) sp_efecto_transforma_0;
            } else if (jugador[i].respawn < TIEMPO_RESPAWN_GIRA1 && jugador[i].respawn > TIEMPO_RESPAWN_REAPARECE) {
                sprite = (u8*) sp_efecto_transforma;
            }

        } else {

            if (!jugador[i].disparando && !jugador[i].herido) {
                if (jugador[i].movimientoV) {
                    sprite = jugador[i].sprite_escala_frame1;
                    if (ciclo > 8) {
                        sprite = jugador[i].sprite_escala_frame2;
                    }
                } else {
                    sprite = jugador[i].sprite_frame[0];
                    if (jugador[i].movimientoH) {
                        if ((ciclo > 5 && ciclo < 10) || ciclo > 15) {
                            sprite = jugador[i].sprite_frame[1];
                        }
                    }
                }
            } else {             
                if (jugador[i].herido) {
                    sprite = jugador[i].sprite_herido;
                    if (jugador[i].herido < TIEMPO_TRANSFORMANDO) {
                        sprite = (u8*) sp_efecto_transforma;
                        if (jugador[i].herido < (TIEMPO_TRANSFORMANDO - (1 * (TIEMPO_TRANSFORMANDO / 3)))) {
                            sprite = (u8*) sp_efecto_transforma_1;
                        } else if (jugador[i].herido < (TIEMPO_TRANSFORMANDO - (2 * (TIEMPO_TRANSFORMANDO / 3)))) {
                            sprite = (u8*) sp_efecto_transforma_0;
                        }
                    } 
                } else {
                    sprite = jugador[i].sprite_disparando;
                }
            }
        }
        // DIBUJAR EL SPRITE
        //
        cpct_drawSpriteMaskedAlignedTable(sprite,
            cpct_getScreenPtr(vmem_start, jugador[i].x, jugador[i].y),
             SP_MAGO0_0_W, SP_MAGO0_0_H, sp_tablaTrans);


        // GUARDAR LA POSICIÓN ANTERIOR PARA PODER BORRARLA (2 posiciones atrás por el tema del doble búfer)
        jugador[i].x_anterior2 = jugador[i].x_anterior;
        jugador[i].y_anterior2 = jugador[i].y_anterior;
        jugador[i].x_anterior = jugador[i].x;
        jugador[i].y_anterior = jugador[i].y;


        // ***** DEBUG **** COORDENADAS DEL JUGADOR 0
        /*
        if (!i) {
            printTxt("    ", 5, 192, NO, 0);
            printTxt("    ", 16, 192, NO, 0);
            printInt(jugador[i].x, 5, 192, NO, 0);
            printInt(jugador[i].y, 16, 192, NO, 0);
        }
        */
        // ***** DEBUG ****
        /*
        // ***** DEBUG **** PARA LAS COORDENADAS DE COLOCACION DE ITEMS
            if (!i && (jugador[i].x % 4) > 0) {
                printTxt("    ", 5, 192, NO, 0);
                printTxt("    ", 16, 192, NO, 0);
            }
            
            if (!i && jugador[i].x % 4 == 0) {
                    printInt(jugador[i].x, 5, 192, NO, 0);
                    printInt(jugador[i].y+5, 16, 192, NO, 0); // Para posicionar ITEMS, sumar 5 a la 'y' del jugador
            }
        // ***** DEBUG ****
        */

    }
    /////

    /////////////////////////////////////////////////////
    // ENTIDADES ITEM
    /////////////////////////////////////////////////////
    // 
    for (i = 0; i < MAX_OBJETOS_ITEM; i++) {
        if (item[i].propiedades & mask_pintar) {
            
            frame = ciclo / 5;
            sprite = item[i].sprite_frame + (frame * SP_ITEM_VIDA_ROTATE_0_W * SP_ITEM_VIDA_ROTATE_0_H) ;
            
            // DIBUJAR EL SPRITE
            cpct_drawSprite (sprite, cpct_getScreenPtr(vmem_start, item[i].x, item[i].y), SP_ITEM_VIDA_ROTATE_0_W, SP_ITEM_VIDA_ROTATE_0_H);
        }
    }
    //

    /////////////////////////////////////////////////////
    // ENTIDADES HECHIZO
    /////////////////////////////////////////////////////
    // 
    for (i = 0; i < MAX_OBJETOS_HECHIZO; i++) {
        if (hechizo[i].tipo != 0) {
            if (!hechizo[i].destruir || hechizo[i].destruir > 2) {
            
                sprite = hechizo[i].sprite_frame1;
                if (hechizo[i].destruir == 0)
                    sprite = hechizo[i].sprite_destruir;
                
                // DIBUJAR EL SPRITE
                //
                cpct_drawSpriteMaskedAlignedTable(sprite,
                        cpct_getScreenPtr(vmem_start, hechizo[i].x, hechizo[i].y),
                        SP_HECHIZO_0_W , SP_HECHIZO_0_H , sp_tablaTrans);

                // GUARDAR LA POSICIÓN ANTERIOR PARA PODER BORRARLA (2 posiciones atrás por el tema del doble búfer)
                hechizo[i].x_anterior2 = hechizo[i].x_anterior;
                hechizo[i].y_anterior2 = hechizo[i].y_anterior;
                hechizo[i].x_anterior = hechizo[i].x;
                hechizo[i].y_anterior = hechizo[i].y;
            }
        }
    }
    //

}

// ==============================================================
// DIBUJAR UN RECUADRO NEGRO CON BORDE BLANCO, COMO UNA VENTANA DE POPUP
//
// - PARÁMETROS:
//      x:  Coordenada X izquierda de inicio del dibujado
//      y:  Coordenada Y superior de inicio del dibujado
//      w:  Longitud horizontal que tendrá el recuadro (en bytes)
//      h:  Longitud vertical que tendrá el recuadro (en pixeles)
// ==============================================================
void pintarRecuadroDoble (u8 x, u8 y, u8 w, u8 h) {
    u8 i = 0;
    u8* vmem_bufer = VMEM_BUFFER1;
    for (i = NUM_BUFFERS; i > 0; i--) {
        cpct_drawSolidBox (cpct_getScreenPtr (vmem_bufer, x , y), 0xFF, w, h);
        cpct_drawSolidBox (cpct_getScreenPtr (vmem_bufer, x + 1 ,y + 2), 0x00, w - 2, h - 4);
        vmem_bufer = VMEM_BUFFER2;
    }
}

